 /*
 * xferstats, a wu-ftpd and ncftpd logfile parser and report generator
 * Copyright 1997  Phil Schwan <pschwan@apk.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 	v1.01 changes:
		* allowed for > 100kb/sec transfer rates in the totals
		* fixed hostname parsing for hostnames without a '.' (ie,
			localhost).  they now show as unresolved
		* made -A the default
		* made parsing compatible with the new way of documenting guest
			logins ('g' instead of 'a')
	v1.02 changes (Sep 27 1997):
		* added -s <section> support
		* made a Makefile for the compiling-impaired
		* dotted quads now show as "unresolved" (as they should!)

	v1.03 changes (Oct 25 1997):
		* check to make sure the xferlog file actually exists
			thanks to: Emil Isberg <emil.isberg@mds.mdh.se>
		* added a small Solaris compatibility flag
			thanks to: Emil Isberg <emil.isberg@mds.mdh.se>
		* added INSTALL file and man page, modified the Makefile

	v1.10 changes (Nov 06 1997):
		* added the "Most Downloaded Files" section
		* added corresponding -p <length> parameter
		* fixed a typo that only affects DEBUG output, but in a big way
		* added more configurable options and split them off into
			config.h
		* boosted variable sizes to accomodate huge xferlogs
		* fixed dotted quad recognition
		* a big big big thanks to Nat Friedman <ndf@mit.edu> and
			Elliot Lee <sopwith@redhat.com> for taking a look at
			the code and sending some key fixes

 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "config.h"

#ifndef SOLARIS
	#include <getopt.h>
#endif

 static char VERSION[] = "xferstats.c  1.10  Copyright (C) Phil Schwan";

const char *MONTHS[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
	"Aug", "Sep", "Oct", "Nov", "Dec", NULL};

static FILE *filePtr;
static char domain[11], onlyDomain[5] = {0}, section[255] = {0}, date[25], userType,
	host[201], fileName[201]={}, lowDate[25] = {0}, highDate[25] = {0}, inOut;
static int depth=4, secs, realTraffic = 1, anonTraffic = 1, hourlyTraffic = 1,
	domainTraffic = 1, totalTraffic = 1, mostDownloads = 1, mostDLNumber =
	DEFAULT_MOSTDL_NUMBER, sectionTraffic = 0, sectionLength = 0,
	outputHTML = DEFAULT_HTML;
static unsigned long long total, hourlyData[24] = {0}, data;
static unsigned int hourlyCount[24] = {0}, hourlySeconds[24] = {0}, fileCount = 0;
struct domains {
	char name[11];
	unsigned long long data;
	unsigned int fileCount;
	unsigned int seconds;
	struct domains *nextPtr;
};
struct days {
	char name[16];
	unsigned long long data;
	unsigned int fileCount;
	unsigned int seconds;
	struct days *nextPtr;
};
struct dirs {
	char name[47];
	unsigned long long data;
	unsigned int fileCount;
	int sorted;
	struct dirs *nextPtr;
};
struct files {
	char name[1024];
	unsigned long long data;
	unsigned int fileCount;
	int sorted;
	struct files *nextPtr;
};
typedef struct domains DomainNode;
typedef DomainNode *NODEPTR;
typedef struct hourStruct hours[24];
typedef struct days DayNode;
typedef DayNode *DAYPTR;
typedef struct dirs DirNode;
typedef DirNode *DIRPTR;
typedef struct files FileNode;
typedef FileNode *FILEPTR;

int parseline(void)
{
	char foo[4096]={}, junkChar;

	if (!feof(filePtr)) {
		fgets(foo, 4096, filePtr);  /* there's probably a better way to do this :) */
#ifdef DEBUG
		printf("RAW: %s", foo);
#endif
		memset(date, 0, sizeof(date));
		strncpy(date, foo, 24);
		sscanf(foo + 25, "%d %s %llu %s %c %c %c %c", &secs, host, &data, fileName, &junkChar, &junkChar, &inOut, &userType);
	} else {
		return 0;
	}
	if (!strcmp(date, ""))  /* this is a hack to make it realize when */
		return 0; 	/* there's an EOF which it doesn't seem to be */
				/* doing correclty. */

#ifdef DEBUG
	printf("%s %d %s %llu %s %c %c\n", date, secs, host, data, fileName, inOut, userType);
#endif
	return 1;
} /* parseline */


void doHours(void)
{
	int hours;
	char foo[3] = {0};

	strncpy(foo, date + 11, 2);
	hours = atoi(foo);
	hourlyData[hours] += data;
	hourlyCount[hours]++;
	hourlySeconds[hours] += secs;
} /* doHours */


int finalDir(DIRPTR *sPtr, char dcat[4096])
{
	DIRPTR newPtr;

	if (*sPtr == NULL) {
		newPtr = malloc(sizeof(DirNode));
		
		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, dcat, 46);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->nextPtr = NULL;
			newPtr->sorted = 0;
			*sPtr = newPtr;
#ifdef DEBUG
			printf("1st directory structure for \"%s\" created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}

	{
		DIRPTR currentPtr = *sPtr;
		char tempDir[47] = {0};

		strncpy(tempDir, dcat, 46);
		while (currentPtr != NULL) {
			if (!strcmp(currentPtr->name, tempDir)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
#ifdef DEBUG
				printf("Added stats to the dir entry \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}
		newPtr = malloc(sizeof(DirNode));
	
		if (newPtr != NULL) {			/* is space available? */
	    		memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, dcat, 46);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->sorted = 0;
			currentPtr = *sPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newPtr;
			newPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New directory structure for \"%s\" created.\n", newPtr->name);
#endif
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}
	return 1;
} /* finalDir */


int finalFile(FILEPTR *fPtr, char fcat[1024])
{
	FILEPTR newFPtr;

	if (*fPtr == NULL) {
		newFPtr = malloc(sizeof(FileNode));
		
		if (newFPtr != NULL) {  /* is space available? */
			memset(newFPtr->name, 0, sizeof(newFPtr->name));
			strncpy(newFPtr->name, fcat, 1023);
			newFPtr->data = data;
			newFPtr->fileCount = 1;
			newFPtr->nextPtr = NULL;
			newFPtr->sorted = 0;
			*fPtr = newFPtr;
#ifdef DEBUG
			printf("1st file structure for \"%s\" created.\n", newFPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}

	{
		FILEPTR currentPtr = *fPtr;

		while (currentPtr != NULL) {
			if (!strcmp(currentPtr->name, fcat)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
#ifdef DEBUG
				printf("Added stats to the file entry \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}
		newFPtr = malloc(sizeof(FileNode));
	
		if (newFPtr != NULL) {			/* is space available? */
	    		memset(newFPtr->name, 0, sizeof(newFPtr->name));
			strncpy(newFPtr->name, fcat, 1023);
			newFPtr->data = data;
			newFPtr->fileCount = 1;
			newFPtr->sorted = 0;
			currentPtr = *fPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newFPtr;
			newFPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New file structure for \"%s\" created.\n", newFPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}
	return 1;
} /* finalFile */


int doDirs(DIRPTR *sPtr, FILEPTR *fPtr)
{
	char **tokens = NULL, *tmpToken, dcat[4096], fcat[1024];
	int i = 2, fuck_me = 1;

	tmpToken = strtok(fileName, "/");
	for(i = 0; tmpToken; i++) {
		tokens = realloc(tokens, sizeof(char *) * (i + 1));
		tokens[i] = strdup(tmpToken);
		tmpToken = strtok(NULL, "/");
	}

	if(!tokens) {	
#ifdef DEBUG
		printf("Bad file name while splitting in doDirs.\n");
#endif
		return 0;
	}

	bzero(dcat, 4096);
	bzero(fcat, 1024);
	for (; fuck_me < i; fuck_me++) {
		if (fuck_me < depth && fuck_me < i - 1) {
			strcat(dcat, "/");
			strcat(dcat, tokens[fuck_me - 1]);
		}
		strcat(fcat, "/");
		strcat(fcat, tokens[fuck_me - 1]);
	}

#ifdef DEBUG
	{
		char dir[47] = {0};

		strncpy(dir, dcat, 46);
		printf("Final dir: \"%s\"\n", dir);
		printf("Final file: \"%s\"\n", fcat);
	}
#endif

	free(tokens);

	if (mostDownloads && inOut == 'o')
		if (!finalFile(fPtr, fcat)) return 0;
	if (!finalDir(sPtr, dcat)) return 0;
	return 1;
} /* doDirs */


int doDaily(DAYPTR *sPtr)
{
	DAYPTR newPtr;
	
	if (*sPtr == NULL) {
		newPtr = malloc(sizeof(DayNode));
		
		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, date, 10);
			strncat(newPtr->name, date + 19, 5);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->nextPtr = NULL;
			newPtr->seconds = secs;
			*sPtr = newPtr;
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	} else {
		DAYPTR currentPtr = *sPtr;
		char tempDay[16] = {0};

		strncpy(tempDay, date, 10);
		strncat(tempDay, date + 19, 5);

		while (currentPtr != NULL) {
#ifdef DEBUG
			printf("currPtr->name: %s\n", currentPtr->name);
#endif
			if (!strcmp(currentPtr->name, tempDay)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
				currentPtr->seconds += secs;
#ifdef DEBUG
				printf("Added stats to the day \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}

	        newPtr = malloc(sizeof(DayNode));
        
       		if (newPtr != NULL) {  /* is space available? */
       			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, date, 10);
			strncat(newPtr->name, date + 19, 5);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->seconds = secs;
			currentPtr = *sPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newPtr;
			newPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New day structure for \"%s\" created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}	
	return 0;
} /* doDaily */
	

int doDomain(NODEPTR *sPtr)
{
	NODEPTR newPtr;
	
	if (*sPtr == NULL) {
		newPtr = malloc(sizeof(DomainNode));
		
		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, domain, 10);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->nextPtr = NULL;
			newPtr->seconds = secs;
			*sPtr = newPtr;
#ifdef DEBUG
			printf("1st domain (\"%s\") created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	} else {
		NODEPTR currentPtr = *sPtr;

#ifdef DEBUG
		printf("DOMAIN: %s\n", domain);
#endif
		while (currentPtr != NULL) {
#ifdef DEBUG
 			printf("cPtr  : %s\n", currentPtr->name);
#endif
			if (!strcmp(currentPtr->name, domain)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
				currentPtr->seconds += secs;
#ifdef DEBUG
				printf("Added to existing domain \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}

	        newPtr = malloc(sizeof(DomainNode));
        
       		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, domain, 10);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->seconds = secs;
			currentPtr = *sPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newPtr;
			newPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New domain structure for \"%s\" created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}	
	return 0;
} /* doDomain */


int compareDates(char date1[25], char date2[25])
{
	float dw1, dw2;
	int i, m = 0;
	char foo[5] = {0};

	strncpy(foo, date1 + 20, 4);
	dw1 = atoi(foo) * 4800;
	strncpy(foo, date2 + 20, 4);
	dw2 = atoi(foo) * 4800;
	strncpy(foo, date1 + 4, 3);
	foo[3] = '\0';
	for (i=1; i <= 12; i++) {
		if (!strcmp(MONTHS[i], foo)) {
			m = i;
			break;
		}
	}
#ifdef DEBUG
	if (!m) printf("ERROR: invalid month\n");
#endif	
	if (m) dw1 += m * 400;

	m = 0;
	strncpy(foo, date2 + 4, 3);
	for (i=1; i <= 12; i++) {
		if (!strcmp(MONTHS[i], foo)) {
			m = i;
			break;
		}
	}
#ifdef DEBUG
	if (!m) printf("ERROR: invalid month\n");
#endif	
	if (m) dw2 += m * 400;

	strncpy(foo, date1 + 8, 2);
	foo[3] = '\0';
	dw1 += atoi(foo);
	strncpy(foo, date2 + 8, 2);
	dw2 += atoi(foo);
	
	if (dw1 == dw2) return 0;
	if (dw1 > dw2)
		return 1;
	else
		return -1;
}	


void displayTotals(NODEPTR *sPtr, DAYPTR *dPtr, DIRPTR *drPtr, FILEPTR *fPtr)
{
	NODEPTR currentPtr;
	DAYPTR currentDayPtr;
	DIRPTR currentDirPtr, hiPtr = NULL;
	FILEPTR currentFilePtr, hiFPtr = NULL;
	int i=0, whee=0;
	char date1[16] = {0}, date2[16] = {0};
	unsigned long long hibyte;

	strncpy(date1, lowDate, 10);
	strncat(date1, lowDate + 19, 5);
	strncpy(date2, highDate, 10);
	strncat(date2, highDate + 19, 5);

	printf("TOTALS FOR SUMMARY PERIOD %s TO %s\n\n", date1, date2);
	printf("Files Transmitted During Summary Period     %11u\n", fileCount);
	printf("Bytes Transmitted During Summary Period     %11llu\n\n", total);
	currentDayPtr = *dPtr;
	for (; currentDayPtr != NULL; i++)
		currentDayPtr = currentDayPtr->nextPtr;
	printf("Average Bytes Transmitted Per File          %11llu\n", total / fileCount);
	printf("Average Files Transmitted Daily             %11u\n", fileCount / i);
	printf("Average Bytes Transmitted Daily             %11llu\n\n", total / i);
	printf("Daily Transmission Statistics\n\n");
	printf("                 Number Of    Number of    Average    Percent Of  Percent Of\n");
	printf("     Date        Files Sent  Bytes  Sent  Xmit  Rate  Files Sent  Bytes Sent\n");
	printf("---------------  ----------  -----------  ----------  ----------  ----------\n");
	currentDayPtr = *dPtr;
	while(currentDayPtr != NULL) {
		printf("%s  %10u  %11llu  %5.1f KB/s  %10.2f  %10.2f\n", currentDayPtr->name, currentDayPtr->fileCount, currentDayPtr->data,
			(float) currentDayPtr->data / currentDayPtr->seconds / 1024, (float) currentDayPtr->fileCount / fileCount * 100, (float) currentDayPtr->data / total * 100);
		currentDayPtr = currentDayPtr->nextPtr;
	}
	if (totalTraffic) {
		printf("%c\nTotal Transfers from each Archive Section (By bytes)\n\n", 12);
		printf("                                                                 ---- %%  Of ----\n");
		printf("                 Archive Section                Files Bytes Sent  Files   Bytes\n");
   		printf(" ---------------------------------------------- ----- ---------- ------- -------\n");
		currentDirPtr = *drPtr;
		for (i = 0; currentDirPtr != NULL; i++)
			currentDirPtr = currentDirPtr->nextPtr;
		while (whee < i) {
			hibyte = 0;
			currentDirPtr = *drPtr;
			while(currentDirPtr != NULL) {
				if (!currentDirPtr->sorted && currentDirPtr->data >= hibyte) {
					hiPtr = currentDirPtr;
					hibyte = currentDirPtr->data;
				}
				currentDirPtr = currentDirPtr->nextPtr;
			}
			whee++;
			hiPtr->sorted = 1;
			printf(" %-46s %5u %10llu %7.2f %7.2f\n", hiPtr->name, hiPtr->fileCount, hibyte, (float) hiPtr->fileCount / fileCount * 100, (float) hibyte / total * 100);
		}
	}
	if (domainTraffic) {
		printf("%c\nTotal Transfer Amount By Domain\n\n", 12);
		printf("             Number Of    Number of     Average    Percent Of  Percent Of\n");
		printf("Domain Name  Files Sent   Bytes Sent   Xmit  Rate  Files Sent  Bytes Sent\n");
		printf("-----------  ----------  ------------  ----------  ----------  ----------\n");
		currentPtr = *sPtr;
		while(currentPtr != NULL) {
			printf("%-11s  %10u  %12llu  %5.1f KB/s  %10.2f  %10.2f\n", currentPtr->name, currentPtr->fileCount, currentPtr->data,
				(float) currentPtr->data / currentPtr->seconds / 1024, (float) currentPtr->fileCount / fileCount * 100, (float) currentPtr->data / total * 100);
			currentPtr = currentPtr->nextPtr;
		}
	}
	printf("\nThese figures only reflect FTP transfers.  There are many sites\n");
	printf("which mount the archives via NFS, and those transfers are not logged\n"); 
	printf("and reported by this program.\n\n");
	if (hourlyTraffic) {
		printf("%c\nHourly Transmission Statistics\n\n", 12);
		printf("                Number Of    Number of     Average    Percent Of  Percent Of\n");
		printf("     Time       Files Sent  Bytes  Sent  Xmit   Rate  Files Sent  Bytes Sent\n");
		printf("--------------  ----------  -----------  -----------  ----------  ----------\n");
		for (i = 0; i <= 23; i++)
			if (hourlyCount[i])
				printf("%-14d  %10u  %11llu  %6.2f KB/s  %10.2f  %10.2f\n", i, hourlyCount[i], hourlyData[i], (float) hourlyData[i] / hourlySeconds[i] / 1024, (float) hourlyCount[i] / fileCount * 100, (float) hourlyData[i] / total * 100);
	}
	if (mostDownloads) {
		whee = 0;
		printf("%c\nMost Downloaded Files (by number)\n\n", 12);
		printf("                                                                ---- %%  Of ----\n");
		printf("                           Filename                       Files  Files   Bytes\n");
		printf(" -------------------------------------------------------- ----- ------- -------\n");
		currentFilePtr = *fPtr;
		for (i = 0; currentFilePtr != NULL; i++)
			currentFilePtr = currentFilePtr->nextPtr;
		while (whee < mostDLNumber) {
			hibyte = 0;
			currentFilePtr = *fPtr; 
			while(currentFilePtr != NULL) {
				if (!currentFilePtr->sorted && currentFilePtr->fileCount >= hibyte) {
					hiFPtr = currentFilePtr;
					hibyte = currentFilePtr->fileCount;
				}
				currentFilePtr = currentFilePtr->nextPtr;
			}
			if (hibyte == 0)
				break;
			whee++;
			hiFPtr->sorted = 1;
			if (strlen(hiFPtr->name) > 56) {
				char tempName[55];
				strcpy(tempName, hiFPtr->name + strlen(hiFPtr->name) - 54); 
				printf("...%-54s %5llu %7.2f %7.2f\n", tempName, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
			} else
				printf(" %-56s %5llu %7.2f %7.2f\n", hiFPtr->name, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
		}
	}
} /* displayTotals */


void displayHTMLTotals(NODEPTR *sPtr, DAYPTR *dPtr, DIRPTR *drPtr, FILEPTR *fPtr)
{
	NODEPTR currentPtr;
	DAYPTR currentDayPtr;
	DIRPTR currentDirPtr, hiPtr = NULL;
	FILEPTR currentFilePtr, hiFPtr = NULL;
	int i=0, whee=0;
	char date1[16] = {0}, date2[16] = {0};
	unsigned long long hibyte;

	strncpy(date1, lowDate, 10);
	strncat(date1, lowDate + 19, 5);
	strncpy(date2, highDate, 10);
	strncat(date2, highDate + 19, 5);

	printf("<HTML><HEAD>\n");
	printf("<TITLE>xferstats: %s to %s</TITLE>\n", date1, date2);
	printf("</HEAD><BODY TEXT=\"#FFFFFF\" BGCOLOR=\"#000000\">\n");

	printf("TOTALS FOR SUMMARY PERIOD %s TO %s<br><br>\n", date1, date2);
	printf("Files Transmitted During Summary Period     %11u<br>\n", fileCount);
	printf("Bytes Transmitted During Summary Period     %11llu<br><br>\n", total);
	currentDayPtr = *dPtr;
	for (; currentDayPtr != NULL; i++)
		currentDayPtr = currentDayPtr->nextPtr;
	printf("Average Bytes Transmitted Per File          %11llu<br>\n", total / fileCount);
	printf("Average Files Transmitted Daily             %11u<br>\n", fileCount / i);
	printf("Average Bytes Transmitted Daily             %11llu<br><br>\n", total / i);
	printf("<H1>Daily Transmission Statistics</H1><br><br>\n");
	printf("<TABLE BORDER=2>\n");
	printf("<TR><TH>Date</TH><TH>Number Of Files Sent</TH><TH>Number of Bytes Sent</TH><TH>Average Xmit Rate</TH><TH>Percent Of Files Sent</TH><TH>Percent Of Bytes Sent</TH></TR>\n");
	currentDayPtr = *dPtr;
	while(currentDayPtr != NULL) {
		printf("<TR><TH>%s</TH><TH>%10u</TH><TH>%11llu</TH><TH>%5.1f KB/s</TH><TH>%10.2f</TH><TH>%10.2f</TH></TR>\n", currentDayPtr->name, currentDayPtr->fileCount, currentDayPtr->data,
			(float) currentDayPtr->data / currentDayPtr->seconds / 1024, (float) currentDayPtr->fileCount / fileCount * 100, (float) currentDayPtr->data / total * 100);
		currentDayPtr = currentDayPtr->nextPtr;
	}
	printf("</TABLE>\n");
	if (totalTraffic) {
		printf("%c\nTotal Transfers from each Archive Section (By bytes)\n\n", 12);
		printf("                                                                 ---- %%  Of ----\n");
		printf("                 Archive Section                Files Bytes Sent  Files   Bytes\n");
   		printf(" ---------------------------------------------- ----- ---------- ------- -------\n");
		currentDirPtr = *drPtr;
		for (i = 0; currentDirPtr != NULL; i++)
			currentDirPtr = currentDirPtr->nextPtr;
		while (whee < i) {
			hibyte = 0;
			currentDirPtr = *drPtr;
			while(currentDirPtr != NULL) {
				if (!currentDirPtr->sorted && currentDirPtr->data >= hibyte) {
					hiPtr = currentDirPtr;
					hibyte = currentDirPtr->data;
				}
				currentDirPtr = currentDirPtr->nextPtr;
			}
			whee++;
			hiPtr->sorted = 1;
			printf(" %-46s %5u %10llu %7.2f %7.2f\n", hiPtr->name, hiPtr->fileCount, hibyte, (float) hiPtr->fileCount / fileCount * 100, (float) hibyte / total * 100);
		}
	}
	if (domainTraffic) {
		printf("%c\nTotal Transfer Amount By Domain\n\n", 12);
		printf("             Number Of    Number of     Average    Percent Of  Percent Of\n");
		printf("Domain Name  Files Sent   Bytes Sent   Xmit  Rate  Files Sent  Bytes Sent\n");
		printf("-----------  ----------  ------------  ----------  ----------  ----------\n");
		currentPtr = *sPtr;
		while(currentPtr != NULL) {
			printf("%-11s  %10u  %12llu  %5.1f KB/s  %10.2f  %10.2f\n", currentPtr->name, currentPtr->fileCount, currentPtr->data,
				(float) currentPtr->data / currentPtr->seconds / 1024, (float) currentPtr->fileCount / fileCount * 100, (float) currentPtr->data / total * 100);
			currentPtr = currentPtr->nextPtr;
		}
	}
	printf("\nThese figures only reflect FTP transfers.  There are many sites\n");
	printf("which mount the archives via NFS, and those transfers are not logged\n"); 
	printf("and reported by this program.\n\n");
	if (hourlyTraffic) {
		printf("%c\nHourly Transmission Statistics\n\n", 12);
		printf("                Number Of    Number of     Average    Percent Of  Percent Of\n");
		printf("     Time       Files Sent  Bytes  Sent  Xmit   Rate  Files Sent  Bytes Sent\n");
		printf("--------------  ----------  -----------  -----------  ----------  ----------\n");
		for (i = 0; i <= 23; i++)
			if (hourlyCount[i])
				printf("%-14d  %10u  %11llu  %6.2f KB/s  %10.2f  %10.2f\n", i, hourlyCount[i], hourlyData[i], (float) hourlyData[i] / hourlySeconds[i] / 1024, (float) hourlyCount[i] / fileCount * 100, (float) hourlyData[i] / total * 100);
	}
	if (mostDownloads) {
		whee = 0;
		printf("%c\nMost Downloaded Files (by number)\n\n", 12);
		printf("                                                                ---- %%  Of ----\n");
		printf("                           Filename                       Files  Files   Bytes\n");
		printf(" -------------------------------------------------------- ----- ------- -------\n");
		currentFilePtr = *fPtr;
		for (i = 0; currentFilePtr != NULL; i++)
			currentFilePtr = currentFilePtr->nextPtr;
		while (whee < mostDLNumber) {
			hibyte = 0;
			currentFilePtr = *fPtr; 
			while(currentFilePtr != NULL) {
				if (!currentFilePtr->sorted && currentFilePtr->fileCount >= hibyte) {
					hiFPtr = currentFilePtr;
					hibyte = currentFilePtr->fileCount;
				}
				currentFilePtr = currentFilePtr->nextPtr;
			}
			if (hibyte == 0)
				break;
			whee++;
			hiFPtr->sorted = 1;
			if (strlen(hiFPtr->name) > 56) {
				char tempName[55];
				strcpy(tempName, hiFPtr->name + strlen(hiFPtr->name) - 54); 
				printf("...%-54s %5llu %7.2f %7.2f\n", tempName, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
			} else
				printf(" %-56s %5llu %7.2f %7.2f\n", hiFPtr->name, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
		}
	}
} /* displayHTMLTotals */


void usage(void)
{
	fprintf(stderr, "%s\n", VERSION);
	fprintf(stderr, "usage: xferstats <options>\n\n"
			" -f <filename>		use <filename> for the log file\n"
			" -H			toggle HTML output\n"
			" -r			include real users\n"
			" -a			include anonymous users\n"
			" -d			include report on domain traffic\n"
			" -h			include report on hourly traffic\n"
			" -t			report on total traffic by section\n" 
			" -m			include a \"most downloaded files\" report\n"
			" -A			include all users, all reports\n"
			" -p <number>		number of \"most-downloaded files\" to display\n"
			" -D <domain>		report only on traffic from <domain>\n"
			" -l <depth>		depth of path detail for sections (default 2)\n"
			" -s <section>		section to report on, for example: -s /pub will\n"
			"			report only on paths under /pub\n"
			" -v			display the version number\n");

} /* usage */


void main(int argc, char **argv)
{
	char *fileStr, ch;
	int realArg = 0;
	NODEPTR startPtr = NULL;
	DAYPTR firstDayPtr = NULL;
	DIRPTR firstDirPtr = NULL;
	FILEPTR firstFilePtr = NULL;

	setbuf(stdout, NULL);
	fileStr = FILENAME;
	while ((ch = getopt(argc, argv, "f:HArahdtvmD:l:s:p:")) != -1)
		switch (ch) {
			case 'f':
				fileStr = optarg;
				break;
			case 'H':
				if (outputHTML) outputHTML = 0;
				else outputHTML = 1;
				break;
			case 'r':
				realArg = 1;
				realTraffic = 2;
				break;
			case 'a':
				realArg = 1;
				anonTraffic = 2;
				break;
			case 'h':
				realArg = 1;
				hourlyTraffic = 2;
				break;
			case 'd':
				realArg = 1;
				domainTraffic = 2;
				break;
			case 't':
				realArg = 1;
				totalTraffic = 2;
				break;
			case 'm':
				realArg = 1;
				mostDownloads = 2;
				break;
			case 'D':
				strncpy(onlyDomain, optarg, 4);
				printf("Transfer totals include the '%s' domain only.\n", onlyDomain);
				printf("All other domains are filtered out for this report.\n\n");
				break;
			case 'p':
				mostDLNumber = atoi(optarg);
				break;
			case 'l':
				depth = atoi(optarg);
				break;
			case 's':
				sectionTraffic = 1;
				if (optarg[0] != '/') {
					strcpy(section, "/");
					strcat(section, optarg);
				} else {
					strncpy(section, optarg, 254);
				}

				sectionLength = strlen(section);
				printf("Transfer totals include the '%s' section only.\n", section);
				printf("All other sections are filtered out for this report.\n\n");
				break;
			case 'v':
				printf("%s\n", VERSION);
				exit(1);
			case 'A':
				realArg = 1;
#ifdef DEFAULT_ANON
				anonTraffic = 2;
#endif
#ifdef DEFAULT_HOURLY
				hourlyTraffic = 2;
#endif
#ifdef DEFAULT_DOMAIN
				domainTraffic = 2;
#endif
#ifdef DEFAULT_TOTALS
				totalTraffic = 2;
#endif
#ifdef DEFAULT_REAL
				realTraffic = 2;
#endif
#ifdef DEFAULT_MOSTDL
				mostDownloads = 2;
#endif
				break;
			default:
				usage();
				exit(1);
		}
	if (realArg) {
		if (anonTraffic < 2) anonTraffic = 0;
		if (hourlyTraffic < 2) hourlyTraffic = 0;
		if (domainTraffic < 2) domainTraffic = 0;
		if (totalTraffic < 2) totalTraffic = 0;
		if (realTraffic < 2) realTraffic = 0;
		if (mostDownloads < 2) mostDownloads = 0;
	}
	filePtr = fopen(fileStr, "r");
        if (filePtr == NULL) {
		perror(fileStr);
		exit(EXIT_FAILURE);
        }
	do {
		if (parseline()) {
			char c[11] = {0}, *d;

#ifdef DEBUG
			printf("%s ftped %s\n", host, fileName);
#endif
			if (strchr(host, '.')) {
				strncpy(c, strrchr(host, '.') + 1, 4);
#ifdef DEBUG
				printf("Last letter of domain: %d:%c\n", c[strlen(c)-1], c[strlen(c)-1]);
#endif
				if (isdigit(c[strlen(c)-1])) strcpy(c, "unresolved");
			} else
				strcpy(c, "unresolved");
			for(d = c; (*d = tolower(*d)); d++);
			strncpy(domain, c, 10);
			domain[10] = '\0';
			if (((!strcmp(onlyDomain, "") || !strcmp(onlyDomain, domain))) && (!sectionTraffic || !strncmp(fileName, section, sectionLength))) {
				total += data;
				fileCount++;
				if ((realTraffic && userType == 'r') || (anonTraffic && (userType == 'a' || userType == 'g'))) {
					if (!doDomain(&startPtr)) {
						printf("fatal: could not doDomain for domain \"%s\".\n", domain);
						exit(2);
					}
					if (!doDaily(&firstDayPtr)) {
						printf("fatal: could not doDaily.\n");
						exit(2);
					}
					if (!doDirs(&firstDirPtr, &firstFilePtr)) {
						printf("fatal: could not doDirs.\n");
						exit(2);
					}
					if (hourlyTraffic) doHours();
					if (!strcmp(lowDate, "")) {
						strncpy(lowDate, date, 24);
						lowDate[24] = '\0';
					} else if (compareDates(date, lowDate) < 0) {
						strncpy(lowDate, date, 24);
						lowDate[24] = '\0';
					}
					if (!strcmp(highDate, "")) {
						strncpy(highDate, date, 24);
						highDate[24] = '\0';
					} else if (compareDates(date, highDate) > 0) {
						strncpy(highDate, date, 24);
						highDate[24] = '\0';
					}
				}
			}
		}
	} while (!feof(filePtr));
	fclose(filePtr);
	if (startPtr != NULL)
		if (outputHTML)
			displayHTMLTotals(&startPtr, &firstDayPtr, &firstDirPtr, &firstFilePtr);
		else
			displayTotals(&startPtr, &firstDayPtr, &firstDirPtr, &firstFilePtr);
	else
		printf("There was no data to process.\n");
}
